#include <cassert>
#include <cstdio>
#include <cmath>
#include <cstdlib>

#include "AuxFuncs.h"
#include "AuxRND.h"
#ifdef __USE_MERSENNE_TWISTER
#include "mtwister.h"
#endif // __USE_MERSENNE_TWISTER

void
ConvertThetaPhiToXYZ(float theta, float phi, float xyz[])
{
  // Here we convert the angles to 3D vector
  xyz[0] = cos(phi) * sin(theta);
  xyz[1] = sin(phi) * sin(theta);
  xyz[2] = cos(theta);
}//--- ConvertThetaPhiToXYZ2 --------------------------------

// The method that converts the directional vector on the hemisphere to the 
// angle theta (from zenith) and phi angle (the azimuth)
void
ConvertXYZtoThetaPhi(float xyz[3], // input direction
		     float &theta, float &phi) // output direction theta and phi
{
  float mag = sqrt(xyz[0] * xyz[0] + xyz[1] * xyz[1]+ xyz[2] * xyz[2]);
  assert(mag > 1e-15f);
  mag = 1.0f/mag - 1e-5f;
  xyz[0] *= mag;
  xyz[1] *= mag;
  xyz[2] *= mag;
  assert(xyz[0] >= -1.f);
  assert(xyz[1] >= -1.f);
  assert(xyz[2] >= -1.f);
  assert(xyz[0] <= 1.f);
  assert(xyz[1] <= 1.f);
  assert(xyz[2] <= 1.f);

  // Standard conversion
  theta = acos(xyz[2]);
  if (theta > M_PI)
    theta = M_PI;
  phi = atan2(xyz[1], xyz[0]);

  if (phi < 0.f)
    phi += 2.f*M_PI;
  if (phi > 2.f * M_PI)
    phi -= 2.f*M_PI;

  assert( (phi >= 0.f) && (phi <= 2.f * M_PI));
  return;
}

void
ArbitraryNormal(const float N[3], float outN[3])
{
  float dist2 = N[0] * N[0] + N[1] * N[1]; // in X and Y
  if (dist2 > 0.001f) { // typical case
    float inv_size = 1.0f/sqrt(dist2);
    outN[0] = N[1] * inv_size;
    outN[1] = -N[0] * inv_size;
    outN[2] = 0.f; // N x (0,0,1)
    return;
  }
  // N is close to the the direction of Z axis
  float inv_size = 1.0/sqrt(N[2] * N[2] + N[0] * N[0]);
  outN[0] = -N[2] * inv_size;
  outN[1] = 0.f;
  outN[2] = N[0] * inv_size; // N x (0,1,0)
  return;
}
