// Written by Vlastimil Havran, 2013

#include <cmath>
#include <cstdlib>
#include <fstream>
#include <cassert>

#include "configg.h"
#include "colorcnv.h"
#include "basstr.h"
#include "brdfmeasured.h"

int dataRead = 0;

#define DIRCAT 0x2f

// destructor
BRDFmeasuredModel::~BRDFmeasuredModel()
{
  delete []vals;
  delete []rgbvals;
}

int
BRDFmeasuredModel::ReadParam(fstream &fs, const char * keyword, int &parVar)
{
  if (FindLineWithKeyword(fs, keyword)) {
    cout << "Error: " << keyword << " not found in file." << endl;
    fs.close();exit(0);
  }
  char ch;
  EatWhitespace(fs); fs >> ch; EatWhitespace(fs);
  fs >> parVar;
  cout << "Parameter: " << keyword << " (value) = " << parVar << endl;
  
  return 0;
}

int
BRDFmeasuredModel::ReadParam(fstream &fs, const char * keyword, float &parVar)
{
  if (FindLineWithKeyword(fs, keyword)) {
    cout << "Error: " << keyword << " not found in file." << endl;
    fs.close();exit(0);
  }
  char ch;
  EatWhitespace(fs); fs >> ch; EatWhitespace(fs);
  fs >> parVar;
  cout << "Parameter: " << keyword << " (value) = " << parVar << endl;
  
  return 0;
}

void
BRDFmeasuredModel::ReadIntensity(fstream &fss, int tli)
{
  float value;
  int i22 = 0;
  // we have to put the values to a new coordinate
  // sytem for theta from 0 to 90 and phi from 0 to 359
  for(int ti=0; ti < thetaCount; ti++) {
    float thetaVal;
    fss >> thetaVal;
    // recompute the theta index from angle value
    //cout << "ThetaVal = " << thetaVal << endl;
    int thetaIndex = GetIndexTheta(thetaVal);
    //cout << "ThetaIndex = " << thetaIndex << endl;
    //cout << "ThetaEnd = " << thetaEnd << endl;
    for(int phi=0; phi < phiCount; phi++) { 
      fss >> value; if (value < 0) value = 0;
      int phiVal = phiBeg + phi * phiStep;
      int phiIndex = GetIndexPhi(phiVal);
#if 0
      float v = getv(phiIndex, thetaIndex, tli);
      if (v > 0) {
	cout << "i22 = " << i22 << " phiIndex = " 
	     << phiIndex << " thetaVal = " << thetaVal
	     << " thetaIlIndex = " << tli << endl;
      }
#endif
      setv(phiIndex, thetaIndex, tli, value);
      //cout << "Value = " << value << endl;
      i22++;
    } // for pi	
  } // for ti
  return;
}

void
BRDFmeasuredModel::ReadSpectrum(fstream &fss, int tli)
{
  float value;
  int i22 = 0;
  // we have to put the values to a new coordinate
  // sytem for theta from 0 to 90 and phi from 0 to 359
  for(int ti=0; ti < thetaCount; ti++) {
    // recompute the theta index from angle value
    //cout << "ThetaVal = " << thetaVal << endl;
    //cout << "ThetaIndex = " << thetaIndex << endl;
    //cout << "ThetaEnd = " << thetaEnd << endl;
    float firstThetaVal;
    for(int phi=0; phi < phiCount; phi++) { 
      float thetaVal;
      fss >> thetaVal;
      int thetaIndex = GetIndexTheta(thetaVal);
      if (phi==0) {
	firstThetaVal = thetaVal;
	//cout << "FirstThetaVal = " << firstThetaVal << endl;
      } else {
	//cout << "ThetaVal = " << thetaVal;
	if (thetaVal != firstThetaVal) {
	  cout << "Problem with reading spectrum file thetaVal = "
	       << thetaVal << " previousTheta = " << firstThetaVal << endl;
	  fss.close();
	  exit(0);
	}
      }
      
      int phiVal = phiBeg + phi * phiStep;
      int phiIndex = GetIndexPhi(phiVal);
      int iii = addr(phiIndex, thetaIndex, tli);
      for (int wi=0; wi < wavelengthCount; wi++) {
	fss >> value; 
	//cout << "Value = " << value << endl;
	if (value < 0) value = 0;
	vals[iii*wavelengthCount + wi] = value;
      } // for wi
    } // for phi
  } // for ti
  return;
}

// Convert spectral measurements to RGB
void
BRDFmeasuredModel::ConvertToRGB()
{
  if (!IsSpectralMeasurement())
    return;
  // it is spectral measurement and can be converted to RGB
  
  // for all indices of lighting
  for (int tli = 0; tli < numberOfFiles; tli++) {
    float value;
    int i22 = 0;
    // we have to put the values to a new coordinate
    // sytem for theta from 0 to 90 and phi from 0 to 359
    for(int ti=0; ti < thetaCount; ti++) {
      // recompute the theta index from angle value
      //cout << "ThetaVal = " << thetaVal << endl;
      //cout << "ThetaIndex = " << thetaIndex << endl;
      //cout << "ThetaEnd = " << thetaEnd << endl;
      for(int phi=0; phi < phiCount; phi++) {       
      // convert to RGB
	float cX, cY, cZ; // color values in XYZ space
	int iii = addr(phi, ti, tli);
#if 0
	float vr = rgbvals[iii];
	float vg = rgbvals[iii+1];
	float vb = rgbvals[iii+2];
	if ((vr > 0) || (vg > 0) || (vb>0)) {
	  cout << "i22 = " << i22 << " phiIndex = " 
	       << phi << " thetaIndex = " << ti
	       << " thetaIlumIndex = " << tli << endl;
	}
#endif
	SpectralToXYZ(cX, cY, cZ,  // output
		      &(vals[iii*wavelengthCount]),
		      wavelengthBeg, wavelengthEnd,
		      wavelengthCount);
	float R, G, B;
	XYZ_to_sRGB(cX, cY, cZ, // input
		    R, G, B); // output
	//cout << R << " " << G << " " << B << " " << flush;
	assert(rgbvals);
	rgbvals[3*iii] = R; // store the colour
	rgbvals[3*iii+1] = G;
	rgbvals[3*iii+2] = B;
	if (R+G+B < 0.) {
	  cout << "Color value is negative\n" << endl;
	  rgbvals[3*iii] = rgbvals[3*iii+1] = rgbvals[3*iii+2] = 0;
	  //abort();
	}
      } // for phi
    } // for ti
  } // for tli
  return;
}

int
BRDFmeasuredModel::ReadBRDFDataFromFile(char *filename)
{
  dataRead = 0; // not yet read from file
  // Here we read the intensity from a file
  string fextension = ExtractFilenameExtension(string(filename));
  ToLowerCase(fextension);
  //cout << "AA " << fextension << endl;
  //cout << fextension.length() << endl;
  if (fextension.compare("ind")==0) {
    // this is the index file to the extensions
    fstream fs;
    fs.open(filename);
    if (!fs)
    {
      cout << "File " << filename << " cannot be opened." << endl;
      fs.close();exit(0);
    }
    modefs = -1;
    ReadParam(fs, "Spectral Mode", modefs);
    if (!((modefs == 0)||(modefs == 1)))  {
      cout << "ERROR - only the mode intensity is supported" << endl;
      cout << "modefs = " << modefs << endl;
      fs.close();exit(-1);
    }
    EatWhitespace(fs);
    if (modefs == 1) {
      ReadParam(fs, "Wavelength number", wavelengthCount);
      ReadParam(fs, "Wavelength begin", wavelengthBeg);
      ReadParam(fs, "Wavelength step", wavelengthStep);
      wavelengthEnd = wavelengthBeg + wavelengthStep*wavelengthCount;
    }
    else { // we only measure intensity
      assert(modefs == 0);
      wavelengthCount = 1;
      wavelengthBeg = 550;
      wavelengthEnd = 550;
      wavelengthStep = 0;
    }

    EatWhitespace(fs);
    ReadParam(fs, "Theta measure beginning", thetaBeg);
    ReadParam(fs, "Theta measure end", thetaEnd);
    ReadParam(fs, "Theta measure step", thetaStep);
    thetaCount = (thetaEnd - thetaBeg)/thetaStep + 1;
    thetaM_Count = thetaCount;
    cout << "thetaCount = " << thetaCount << endl;
    int scanV;
    ReadParam(fs, "Scanning", scanV);    
    if (FindLineWithKeyword(fs, " axes")) {
      cout << "The keyword for axes was not found" << endl;
      fs.close();exit(0);
    }
    ReadParam(fs, "Phi measure beginning", phiBeg);
    ReadParam(fs, "Phi measure end", phiEnd);
    phiEndMax = phiEnd + phiEnd - phiBeg;
    ReadParam(fs, "Phi measure step", phiStep);
    phiCount = (phiEnd - phiBeg)/phiStep + 1;
    phiM_Count = phiCount;
    cout << "phiCount = " << phiCount << endl;
    ReadParam(fs, "Theta Lighting beginning", thetaLightingBeg);
    ReadParam(fs, "Theta Lighting end", thetaLightingEnd);
    ReadParam(fs, "Theta Lighting step", thetaLightingStep);
    thetaLightingCount = (thetaLightingEnd - thetaLightingBeg)
      /thetaLightingStep+1;
    thetaLightingM_Count = thetaLightingCount;
    cout << "ThetaLightingCount = " << thetaLightingCount << endl;
    ReadParam(fs, "Number of files", numberOfFiles);
    assert(thetaLightingCount == numberOfFiles);
    EatWhitespace(fs);
    // allocate the space for values
    if (modefs == 0)
      maxvindex = thetaCount * phiCount * thetaLightingCount;
    else {
      maxvindex = thetaCount * phiCount * thetaLightingCount;
      // Check if it fits into integer
      int intMax = (int)(((1L<<31)-1L));
      int cntMax = intMax / maxvindex;
      if (cntMax < wavelengthCount) {
	cout << "The number of measurement does not fit into integer" << endl;
	cout << "cntMax = " << cntMax << " wavelengthCount = " 
	     << wavelengthCount << " maxInteger = " << intMax << endl;
	exit(0);
      }
      maxvrgbindex = maxvindex * 3;
      maxvindex *= wavelengthCount;
      rgbvals = new float[maxvrgbindex];
      assert(rgbvals);
      for (int i = 0; i < maxvrgbindex; i++) 
	rgbvals[i] = 0; // for debugging
    }
    vals = new float[maxvindex];
    assert(vals);
    // setting the initial value
    for (int i = 0; i < maxvindex; i++) 
      vals[i] = 0; // for debugging

    // for all indices of lighting
    for (int tli = 0; tli < numberOfFiles; tli++) {
      char filename2[256];
      for (int j = 0; j < 256; j++) filename2[j] = 0;
      fs.getline(filename2, 256);
      string fsname;
      ChangeFilenameExtension(string(filename2), string("txt"), fsname);
      //cout << "i = " << tli << " filename = " << filename2 << endl;
      fstream fss;
      fss.open(fsname.c_str());
      if (!fss)
      {
	cout << "File " << fsname << " cannot be opened." << endl;
	//for (int j = 0; j < 256; j++) {
	//  if (filename2[j] == 0) break;
	//  cout << hex << (int)filename2[j] << " ";
	//}
	//cout << endl;
	fss.close();exit(0);
      }
      else
	cout << "File " << fsname << " was opened." << endl;
      // Read the data from each individual file
      if (FindLineWithKeyword(fss, "Theta Lighting")) {
	cout << "The keyword for Theta Lighting was not found" << endl;
	fss.close();exit(0);
      }
      EatWhitespace(fss);
      float thetaLight;
      fss >> thetaLight;
      cout << "ThetaLight = " << thetaLight << endl;
      EatWhitespace(fss);
      if (modefs == 0)
	ReadIntensity(fss, tli);
      if (modefs == 1)
	ReadSpectrum(fss, tli);
      fss.close();
    } // for i
    fs.close();
    int negCount = 0;
    for (int i = 0; i < maxvindex; i++) {
      if (vals[i] < -99) negCount++;
    } 
    cout << "negCount = " << negCount 
	 << " maxvindex " << maxvindex << endl;    
  }
  else {
    printf("File .ind has to be given\n");
    exit(0);
  }
  
  dataRead = 1; // data were read
  return 0; // no problem
}

// This uses the measured data from the file
void
BRDFmeasuredModel::GetBRDF(float RGB[3], // output RGB
			   float cosangle, //
			   float refl_illum_vec[3], // illumination vector
			   float view_vec[3]) const // view vector
{
  // compute the four values around and compute bilinear interpolation
  cout << "Not yet implemented" << endl;
}

void
BRDFmeasuredModel::GetBRDF(float RGB[3], // output RGB
			   float theta_illum, // illumination vector
			   float phi_illum,
			   float theta_view, // view vector
			   float phi_view) const
{
  // all the angles are specified in degrees, phi in range <0,360)
  // and theta in range <0,90)
#if 0
  cout << " ti = " << theta_illum
       << " pi = " << phi_illum
       << " tv = " << theta_view
       << " pv = " << phi_view << endl;
#endif
  
  if (theta_illum > ThetaLightingEnd())
    theta_illum = ThetaLightingEnd();

  int til = GetIndexThetaLighting(theta_illum);
  int til2 = til+1;
  //cout << "til = " << til << " til2 = " << til2 << endl;
  float tilw = (float)(theta_illum - GetValThetaLighting(til))/thetaLightingStep;
  if (til2 >= thetaLightingCount) til2 = til;

  // Since it is isotropic 3D function
  float phiD = (phi_illum - phi_view);
  //cout << phiD << " " << endl;
  while (phiD < phiBeg) phiD += 360;
  while (phiD > phiEndMax) phiD -= 360;
  assert(phiD >= phiBeg);
  assert(phiD <= phiEndMax);

#if 0
  if (fabs(phiD - phiEnd) < 2) { // degrees
    RGB[0] = RGB[1] = RGB[2] = 0;
    return;
  }
#endif
  bool revertFlag = false;  

  // Now we have it in the range <phiBeg, phiEndMax>
  if (phiD > phiEnd) {
    //phiD = 2*phiEnd - phiD;
    revertFlag = true;
    phiD = phiD - phiEnd + phiBeg;
    theta_view = -theta_view;
    //cout << "theta: " << theta_view << " " << thetaBeg << " " << thetaEnd << endl;
  }
  if (theta_view > 90.f) {
    RGB[0] = RGB[1] = RGB[2] = 0.f;
    return;
  }
  if (theta_view < -90.f) {
    RGB[0] = RGB[1] = RGB[2] = 0.f;
    return;
  }
  //assert(theta_view >= thetaBeg);
  //assert(theta_view <= thetaEnd);
  
  // phi
  int piv = GetIndexPhi(phiD);
  int piv2 = piv+1;
  if (piv2 >= phiCount) {
    //cout << "SS" << endl;
    piv2 = 0;
  }
  //cout << "piv = " << piv << " piv2 = " << piv2 << endl;
  float pivw = (float)(phiD - GetValPhi(piv))/phiStep;
  if (piv2 >= phiCount)
    piv2 = piv; // limitation
  

  // Finally compute theta for view
  int tiv = GetIndexTheta(theta_view);
  //cout << "tiv = " << tiv << " tiv2 = " << tiv2 << endl;
  float tivw = (float)(theta_view - GetValTheta(tiv))/thetaStep;
  if (revertFlag) {
    tivw = 1.0 - tivw;
    //tiv = tiv-1;
    //if (tiv < 0) tiv = 0;
  }

  int tiv2 = tiv+1;
  if (tiv2 >= thetaCount) tiv2 = tiv;
  
  if (modefs == 0) {
    // compute the four values around and compute trilinear interpolation
    float RGB000 = getv(piv, tiv, til);
    assert(RGB000 >= 0);
    float RGB001 = getv(piv, tiv, til2);
    assert(RGB001 >= 0);
    float RGB010 = getv(piv, tiv2, til);
    assert(RGB010 >= 0);
    float RGB011 = getv(piv, tiv2, til2);
    assert(RGB011 >= 0);
    float RGB100 = getv(piv2, tiv, til);
    assert(RGB100 >= 0);
    float RGB101 = getv(piv2, tiv, til2);
    assert(RGB101 >= 0);
    float RGB110 = getv(piv2, tiv2, til);
    assert(RGB110 >= 0);
    float RGB111 = getv(piv2, tiv2, til2);
    assert(RGB111 >= 0);

    // only theta_illum and phi_view are interpolated
    float val1 = RGB000 * (1.0-tivw) * (1.0 - tilw) +
      RGB001 * (1.0-tivw) * tilw +
      RGB010 * tivw * (1.0 - tilw) +
      RGB011 * tivw * tilw;
    float val2 = RGB100 * (1.0-tivw) * (1.0 - tilw) +
      RGB101 * (1.0-tivw) * tilw +
      RGB110 * tivw * (1.0 - tilw) +
      RGB111 * tivw * tilw;
    
    RGB[0] = RGB[1] = RGB[2] = 0.000004*(val1 * (1.0-pivw) + val2 * pivw);
    return;
  }
  // computing for all 3 colour channels
  {
    // compute the four values around and compute trilinear interpolation
    float RGB000[3]; getv(piv, tiv, til, RGB000);
    assert(RGB000[0]+RGB000[1]+RGB000[2] >= 0);
    float RGB001[3]; getv(piv, tiv, til2, RGB001);
    assert(RGB001[0]+RGB001[1]+RGB001[2] >= 0);
    float RGB010[3]; getv(piv, tiv2, til, RGB010);
    assert(RGB010[0]+RGB010[1]+RGB010[2] >= 0);
    float RGB011[3]; getv(piv, tiv2, til2, RGB011);
    assert(RGB011[0]+RGB011[1]+RGB011[2] >= 0);
    float RGB100[3]; getv(piv2, tiv, til, RGB100);
    assert(RGB100[0]+RGB100[1]+RGB100[2] >= 0);
    float RGB101[3]; getv(piv2, tiv, til2, RGB101);
    assert(RGB101[0]+RGB101[1]+RGB101[2] >= 0);
    float RGB110[3]; getv(piv2, tiv2, til, RGB110);
    assert(RGB110[0]+RGB110[1]+RGB110[2] >= 0);
    float RGB111[3]; getv(piv2, tiv2, til2, RGB111);
    assert(RGB111[0]+RGB111[1]+RGB111[2] >= 0);

    // only theta_illum and phi_view are interpolated
    for (int i = 0; i < 3; i++) {
      float val1 = RGB000[i] * (1.0-tivw) * (1.0 - tilw) +
	RGB001[i] * (1.0-tivw) * tilw +
	RGB010[i] * tivw * (1.0 - tilw) +
	RGB011[i] * tivw * tilw;
      float val2 = RGB100[i] * (1.0-tivw) * (1.0 - tilw) +
	RGB101[i] * (1.0-tivw) * tilw +
	RGB110[i] * tivw * (1.0 - tilw) +
	RGB111[i] * tivw * tilw;    
      RGB[i] = 1.0 * (val1 * (1.0-pivw) + val2*pivw);
    }
  }
  return;
}

// ----------------------------------------------------------------------------

BRDFmeasuredModelComposed::~BRDFmeasuredModelComposed()
{
  delete brdf2;
}

int
BRDFmeasuredModelComposed::ReadBRDFDataFromTwoFiles(char *filename1,
						    char *filename2)
{
  ReadBRDFDataFromFile(filename1);
  brdf2 = new BRDFmeasuredModel;
  assert(brdf2);
  if (brdf2->ReadBRDFDataFromFile(filename2)) {
    cout << "Data from " << filename2 << " cannot be read" << endl;
    cout << "Exiting." << endl;
    abort();
  }
  if ( IsSpectralMeasurement() != brdf2->IsSpectralMeasurement()) {
    cout << "ERROR: Spectral Measurement for two functions do not fit" << endl;
    abort();
  }
  if ( ThetaLightingBeg() != brdf2->ThetaLightingBeg()) {
    cout << "ERROR: ThetaLightingBeg() for two functions do not fit" << endl;
    abort();
  }
  if ( ThetaLightingEnd() != brdf2->ThetaLightingEnd()) {
    cout << "ERROR: ThetaLightingEnd() for two functions do not fit" << endl;
    abort();
  }
  if ( ThetaLightingCount() != brdf2->ThetaLightingCount()) {
    cout << "ERROR: ThetaLightingCount() for two functions do not fit" << endl;
    abort();
  }
  if ( ThetaBeg() != brdf2->ThetaBeg()) {
    cout << "ERROR: ThetaBeg() for two functions do not fit" << endl;
    abort();
  }
  if ( ThetaEnd() != brdf2->ThetaEnd()) {
    cout << "ERROR: ThetaEnd() for two functions do not fit" << endl;
    abort();
  }
  if ( ThetaCount() != brdf2->ThetaCount()) {
    cout << "ERROR: ThetaCount() for two functions do not fit" << endl;
    abort();
  }
  if ( PhiBeg() != brdf2->PhiBeg()) {
    cout << "ERROR: PhiBeg() for two functions do not fit" << endl;
    abort();
  }
  if ( PhiEnd() != brdf2->PhiEnd()) {
    cout << "ERROR: PhiEnd() for two functions do not fit" << endl;
    abort();
  }
  if ( PhiCount() != brdf2->PhiCount()) {
    cout << "ERROR: PhiCount() for two functions do not fit" << endl;
    abort();
  }
  // -------------------------------
  if (IsSpectralMeasurement()) {
    if ( fabs(WavelengthBeg() - brdf2->WavelengthBeg()) > 0.1) {
      cout << "ERROR: WavelengthBeg() for two functions do not fit" << endl;
      cout << "First beg = " << WavelengthBeg() << endl;
      cout << "Second beg = " << brdf2->WavelengthBeg() << endl;
      abort();
    }
    if ( fabs(WavelengthEnd() - brdf2->WavelengthEnd()) > 0.1) {
      cout << "ERROR: WavelengthEnd() for two functions do not fit" << endl;
      abort();
    }
    if ( WavelengthCount() != brdf2->WavelengthCount()) {
      cout << "ERROR: WavelengthCount() for two functions do not fit" << endl;
      abort();
    }    
  }

  // Now we have to substract two files or do other operation and
  // convert again to RGB!
  for (int i = 0; i < maxvindex; i++) {
    vals[i] -= brdf2->GetVal(i); // run the operation
  }

  return 0;
}
